package de.fhdw.wtf.parser.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.util.Vector;

import org.junit.Before;
import org.junit.Test;

import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.GroupElement;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.QualifiedName;
import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException;
import de.fhdw.wtf.common.stream.SimpleTokenStream;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.EndToken;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.common.token.keywords.GroupToken;
import de.fhdw.wtf.common.token.symbols.ColonToken;
import de.fhdw.wtf.common.token.symbols.EqualToken;
import de.fhdw.wtf.common.token.symbols.SemicolonToken;
import de.fhdw.wtf.common.token.symbols.SquareBracketCloseToken;
import de.fhdw.wtf.common.token.symbols.SquareBracketOpenToken;
import de.fhdw.wtf.parser.Parser;

/**
 * Tests the parsing of GroupTokens.
 */
public class TestGroups {
	
	private TokenStream stream;
	
	private IdentifierToken idGroup1Token;
	private IdentifierToken idGroup2Token;
	private IdentifierToken idGroup3Token;
	
	private ColonToken colonToken;
	private GroupToken groupToken;
	private EqualToken equalToken;
	private SquareBracketOpenToken squareBracketOpenToken;
	private SquareBracketCloseToken squareBracketCloseToken;
	private SemicolonToken semicolonToken;
	private EndToken endToken;
	
	private Parser parser;
	
	@Before
	public void setUp() throws Exception {
		this.stream = SimpleTokenStream.create();
		this.idGroup1Token = IdentifierToken.create("Group1", Position.create("", 0 + 1, 0 + 1, 0));
		this.idGroup2Token = IdentifierToken.create("Group2", Position.create("", 0 + 1, 0 + 1, 0));
		this.idGroup3Token = IdentifierToken.create("Group3", Position.create("", 0 + 1, 0 + 1, 0));
		this.colonToken = ColonToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.groupToken = GroupToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.equalToken = EqualToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.semicolonToken = SemicolonToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.parser = Parser.create(this.stream);
		this.endToken = EndToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.squareBracketCloseToken = SquareBracketCloseToken.create(Position.create("", 0 + 1, 0 + 1, 0));
		this.squareBracketOpenToken = SquareBracketOpenToken.create(Position.create("", 0 + 1, 0 + 1, 0));
	}
	
	@Test
	public void testOnlyEndToken() throws Exception {
		this.stream.add(this.endToken);
		
		final Model model = Model.create(this.endToken);
		
		final Model result = this.parser.parse();
		assertEquals(model, result);
		assertEquals(0, this.parser.getExceptions().size());
	}
	
	/**
	 * Group1:group=[]; .
	 */
	@Test
	public void testSimpleGroupComplete() throws Exception {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		final Vector<GroupElement> group1Vector = new Vector<>();
		final Name group1Name = UnqualifiedName.create(this.idGroup1Token);
		final Group group1 = Group.create(group1Name, group1Vector, this.idGroup1Token);
		final Model expected = Model.create(this.idGroup1Token);
		expected.addGroup(group1);
		
		final Model actual = this.parser.parse();
		assertEquals(expected.getGroups().size(), actual.getGroups().size());
		assertEquals(0, this.parser.getExceptions().size());
		assertEquals(expected, actual);
	}
	
	/**
	 * Group1:group=[] .
	 */
	@Test
	public void testSimpleGroupWithoutSemicolon() {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		try {
			this.parser.parse();
			fail();
		} catch (final NoValidTokenStreamException e) {
			assertEquals(1, this.parser.getExceptions().size());
		}
	}
	
	/**
	 * Group1:group=[; .
	 */
	@Test
	public void testSimpleGroupWithoutClose() {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		try {
			this.parser.parse();
		} catch (final NoValidTokenStreamException e) {
			assertEquals(1, this.parser.getExceptions().size());
		}
	}
	
	/**
	 * Group1:group=[Group2:group=[];]; .
	 */
	@Test
	public void testSimpleGroupInGroupComplete() throws Exception {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup2Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		final Model model = Model.create(this.idGroup1Token);
		final Vector<GroupElement> group1Vector = new Vector<>();
		final Vector<GroupElement> group2Vector = new Vector<>();
		final UnqualifiedName group1Name = UnqualifiedName.create(this.idGroup1Token);
		final QualifiedName group2Name = QualifiedName.create(group1Name, UnqualifiedName.create(this.idGroup2Token));
		final Group group1 = Group.create(group1Name, group1Vector, this.idGroup1Token);
		final Group group2 = Group.create(group2Name, group2Vector, this.idGroup2Token);
		group1Vector.add(group2);
		model.addGroup(group1);
		
		final Model result = this.parser.parse();
		assertEquals(model.getGroups().size(), result.getGroups().size());
		assertEquals(0, this.parser.getExceptions().size());
		assertEquals(model, result);
	}
	
	/**
	 * Group1:group=[Group2:group=[;; .
	 */
	@Test
	public void testSimpleGroupInGroup2GroupCloseMissing() {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup2Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		try {
			this.parser.parse();
			fail();
		} catch (final NoValidTokenStreamException e) {
			System.out.println(this.parser.getExceptions());
			assertEquals(3, this.parser.getExceptions().size());
		}
	}
	
	/**
	 * Group1:group=[Group2:group=[group3:group=[];;; .
	 */
	@Test
	public void testSimpleGroupInGroupInGroup2GroupCloseMissing() {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup2Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup3Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		try {
			this.parser.parse();
			fail();
		} catch (final NoValidTokenStreamException e) {
			assertEquals(2, this.parser.getExceptions().size());
		}
	}
	
	/**
	 * "Group1:group=[Group2:group=[group3:group=[];];];" .
	 */
	@Test
	public void testSimpleGroupInGroupInGroup2Group() throws Exception {
		this.stream.add(this.idGroup1Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup2Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.idGroup3Token);
		this.stream.add(this.colonToken);
		this.stream.add(this.groupToken);
		this.stream.add(this.equalToken);
		this.stream.add(this.squareBracketOpenToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(this.squareBracketCloseToken);
		this.stream.add(this.semicolonToken);
		this.stream.add(EndToken.create(Position.create("", 0 + 1, 0 + 1, 0)));
		
		final Model expected = Model.create(this.idGroup1Token);
		final Vector<GroupElement> group1Vector = new Vector<>();
		final Vector<GroupElement> group2Vector = new Vector<>();
		final Vector<GroupElement> group3Vector = new Vector<>();
		final UnqualifiedName group1Name = UnqualifiedName.create(this.idGroup1Token);
		final QualifiedName group2Name = QualifiedName.create(group1Name, UnqualifiedName.create(this.idGroup2Token));
		final Name group3Name = group2Name.addName(this.idGroup3Token);
		final Group group1 = Group.create(group1Name, group1Vector, this.idGroup1Token);
		final Group group2 = Group.create(group2Name, group2Vector, this.idGroup2Token);
		final Group group3 = Group.create(group3Name, group3Vector, this.idGroup3Token);
		group1Vector.add(group2);
		group2Vector.add(group3);
		expected.addGroup(group1);
		
		final Model actual = this.parser.parse();
		assertEquals(expected, actual);
	}
}
